/*************************************************************************/
/*
 *  Datei           : calc.cpp
 *  Inhalt          : Definition fuer Klassen
 *                          o Calculator
 *                        o MyYYScan
 *  Programmierer   : Sven Goethel
 *                      Stapenhorststr.35a
 *                      33615 Bielefeld
 *                      0521/139228
 *  Erstellungsdatum: 06.07.1994
 *  nderungsdatum  : 22.08.1994
 *  Projekt         : calc
 *  Verzeichnis     :
 *  Dokumentation   : calc.doc
 *
 *    (C) Copyright Dr.-Ing.F.Prekwinkel GmbH, 32052 Herford, Germany
 *
 * $Log: calc.cc  $
 * Revision 2.26  1995/06/06 08:48:20  Sven_Goethel_DFP
 * ok
 * Revision 2.25  1995/06/06 08:20:18  Sven_Goethel_DFP
 * lokale char Buffer mit fester Groesse -> lokale StrClass (dynamische groesse)
 * Revision 2.24  1995/04/07 11:34:59  Sven_Goethel_DFP
 * t_string wieder zurueck zu StrClass
 * Alles angepasst fuer Horts
 * ( aehh Horst . )
 * Wenn __SHOW_CALC__ definiert ist,
 * so werden Zeichen auf stderr gedruckt bei :
 *   PARSEN der DATEI
 *   INTERPRETIEREN der DATENSTRUKTUR
 * und alles fuer Calculator NEU und OLD
 * Revision 2.22  1995/04/05 07:45:32  Sven_Goethel_DFP
 * Alle Fehler behoben ..
 * getestet mit und ohne __OLD_INTERPRETE__ !
 * Fuer den neuen Interpreter sind Fehler-Message (FILE und ZEILENNUMMER)
 * hinzugekommen.
 * Fuer __OLD_INTERPRETE__ wurde ein BUG entfernt :
 *   bei return innerhalb verschachtelter Bloecke wurde der 
 *   Block zaehler nicht wieder zureuckgesetzt !!!
 * Revision 2.21  1995/04/05 02:00:01  Sven_Goethel_DFP
 * profekt-pflege nun mit KornShell-Sripts
 * Revision 2.2  1995/04/05 01:48:00  Sven_Goethel_DFP
 * __Mem_Free fehler beseitigt
 * Das MKS System hat um ein mittels new allokiertes Array zu deallokieren,
 * den delete operator verwendet.
 * Es muss aber der delete[] operator verwendet werden :
 *   int *p=new int[100];
 *   delete p; // Error !
 *   delete[] p; // Ok !
 * Revision 2.1  1995/04/01 02:41:34  Sven_Goethel_DFP
 * Referenz fuer Globale Variablen beseitigt, da FEHLER !!!!
 * Das Watcom-Executable stuerzt ab, wenn mehr als zwei ELSE zweige
 * existieren ( oder schon ab zwei ??) <- im test mit test.clc zu ersehen.
 * Das Borland C++ 4.0 executable hingegen arbeitet fehlerfrei !!!
 * Revision 1.6  1995/02/28 15:07:53  Sven_Goethel_WPP
 * source ist nun mit borland c++ version 4.0
 * und mit watcom c++ version 10.0 zu uebersetzen !!!!
 * Revision 1.4  1995/01/30 08:18:04  Sven_Goethel
 * 1st full implemented & checked version
 * Revision 1.3  1995/01/27 17:45:14  Sven_Goethel
 * break - 
 * Revision 1.2  1994/12/16 16:53:31  Sven_Goethel
 * *** empty log message ***
 * Revision 1.1  1994/12/16 16:51:54  Sven_Goethel
 * Initial revision
 *
 */
/*************************************************************************/

#include "calc.h"

static MyYYScan<yy_scan> Scanner(stdin);

const char *Calculator::FirstPrgLine="Calculator ( EXPRESSION );\n";
const char *Calculator::LastPrgLine="\nCalculatorEnd;\n";

const char *Calculator::ExpressionFileName="EXPRESSN.CLC";
const char *Calculator::ExpressionVarName="__EXPRESSION__";
const char *Calculator::ErrorStrings[Calculator::WRONG_ERR_NUMBER+1]= {
                            "No Error detected",
                            /* -------------------------------------*/
                            "Error while getting a variable",
                            "Error while setting a variable",
                            "Variable type does not fit",
                            "Error while allocating memory",
                            "Error while declaring a variable",
                            "Error while interpreting",
                            "Error while opening the Expression-File",
                            /* -------------------------------------*/
                            "Wrong Error Number"
                       };

/*------------------------------------------------------------------------------

    Konstruktor der Klasse Calculator
    --------------------------------------------
    ::Calculator()

    - Beschreibung
    --------------
    Konstruiert eine Instanz des Calculator.
    Die Variablen werden Zurueckgesetzt Reset.

    - Parameter
    -----------
    E: fCalc    - die Eingabedatei

    - allgemeine Ein- und Ausgabeeinheiten
    --------------------------------------
    dieselben des Parsers.


    - Returncode
    ------------
    keiner

    - Historie
    ----------
    6.7.1994 Sven Goethel

------------------------------------------------------------------------------*/
Calculator::Calculator ()
: ErrorState(NO_ERR), yy_parse(NULL, 600)
{
  #ifndef __OLD_INTERPRETE__
    LastBufferPtr=NULL;
  #endif
  Reset();
}

/*------------------------------------------------------------------------------

    Konstruktor der Klasse Calculator
    --------------------------------------------
    void ::Declare(const char *Expression)

    - Beschreibung
    --------------
    Deklarierung einer Variable mittels String-Ausdruckes

    - Parameter
    -----------
    E: Expression    - Der Ausdruck

    - allgemeine Ein- und Ausgabeeinheiten
    --------------------------------------
    dieselben des Parsers.


    - Returncode
    ------------
    keiner

    - Historie
    ----------
    6.7.1994 Sven Goethel

------------------------------------------------------------------------------*/
void Calculator::Declare(const char *Expression)
{
  if( Expression!=NULL && strlen(Expression)>0  )
  {
    StrClass Buffer=(StrClass)FirstPrgLine+(StrClass)Expression
                    +(StrClass)";\n"+(StrClass)";\n"+(StrClass)LastPrgLine;
    #ifndef __OLD_INTERPRETE__
      // Retten des geparsten !!!
      const char * save = LastBufferPtr;
      Anweisungen * pAnweisungen = ObjInhalt.pAnweisungen;
      // Leere Struktur Simulieren ...
      LastBufferPtr=NULL;
      ObjInhalt.pAnweisungen=NULL;
      ObjInhalt.funktionen.FunktionenOff=1;
    #endif
      SetInputFileName("Calculator Buffer\n");
      StartParser((const char *)Buffer); // ErrorState wird implizit gesetzt !!!
    #ifndef __OLD_INTERPRETE__
      ObjInhalt.Do();
      LastBufferPtr=save;
      ObjInhalt.pAnweisungen=pAnweisungen;
      ObjInhalt.funktionen.FunktionenOff=0;
    #endif
  }
}

/*------------------------------------------------------------------------------

    Konstruktor der Klasse Calculator
    --------------------------------------------
    void ::DeclareReal(const char *VarName, double Wert)

    - Beschreibung
    --------------
    Deklarierung einer Variable mittels VariablenNamen und
    optionalen Wert

    - Parameter
    -----------
    E: VarName    - Variablen Name
    E: Wert        - Variablen Wert

    - allgemeine Ein- und Ausgabeeinheiten
    --------------------------------------
    dieselben des Parsers.


    - Returncode
    ------------
    keiner

    - Historie
    ----------
    6.7.1994 Sven Goethel

------------------------------------------------------------------------------*/
void Calculator::DeclareReal(const char *VarName, double Wert)
{
  if( VarName!=NULL && strlen(VarName)>0  )
  {
    VariablenInhalt VarInhalt(0, REELL, Wert);
    if( Vars.DefiniereVar ( VarName , VarInhalt) == 0 )
      ErrorState=VAR_DECLARE_ERR;
  }
}

/*------------------------------------------------------------------------------

    Konstruktor der Klasse Calculator
    --------------------------------------------
    void ::SetReal(const char *VarName, double Wert)

    - Beschreibung
    --------------
    Setzen eines Variablen Wertes mittels VariablenNamen

    - Parameter
    -----------
    E: VarName    - Variablen Name
    E: Wert        - Variablen Wert

    - allgemeine Ein- und Ausgabeeinheiten
    --------------------------------------
    dieselben des Parsers.


    - Returncode
    ------------
    keiner

    - Historie
    ----------
    6.7.1994 Sven Goethel

------------------------------------------------------------------------------*/
void Calculator::SetReal(const char *VarName, double Wert)
{
  if( VarName!=NULL && strlen(VarName)>0  )
  {
    VariablenInhalt VarInhalt(0, REELL, Wert);
    if( Vars.SetzeWert ( 0, VarName , VarInhalt) == 0 )
      ErrorState=VAR_SET_ERR;
  }
}

/*------------------------------------------------------------------------------

    Konstruktor der Klasse Calculator
    --------------------------------------------
    double ::GetReal(const char *VarName)

    - Beschreibung
    --------------
    Holen eines Variablen Wertes mittels VariablenNamen

    - Parameter
    -----------
    E: VarName    - Variablen Name

    - allgemeine Ein- und Ausgabeeinheiten
    --------------------------------------
    dieselben des Parsers.


    - Returncode
    ------------
    Variablen Wert

    - Historie
    ----------
    6.7.1994 Sven Goethel

------------------------------------------------------------------------------*/
double Calculator::GetReal(const char *VarName)
{
  VariablenInhalt VarInhalt;
  double result=0;

  if( VarName!=NULL && strlen(VarName)>0  )
  {
    if(Vars.HoleWert(0, VarName, VarInhalt)==0)
    {
      ErrorState=VAR_GET_ERR;
    } else {
      if(VarInhalt.Type!=REELL) ErrorState=VAR_TYP_ERR;
      else result=VarInhalt.rInhalt;
    }
  }
  return result;
}

/*------------------------------------------------------------------------------

    Konstruktor der Klasse Calculator
    --------------------------------------------
    double ::GetRealExpr(const char *Expression)

    - Beschreibung
    --------------
    Gibt den berechneten Mathematischen Ausdruck zurueck.

    - Parameter
    -----------
    E: Expression    - Mathematischer Ausdruck

    - allgemeine Ein- und Ausgabeeinheiten
    --------------------------------------
    dieselben des Parsers.


    - Returncode
    ------------
    Ergebniss des Ausdrucks

    - Historie
    ----------
    21.7.1994 Sven Goethel

------------------------------------------------------------------------------*/
double Calculator::GetRealExpr(const char *Expression)
{
  StrClass Buffer;
  double result=0;

  if( Expression!=NULL && strlen(Expression)>0  )
  {
    Buffer=(StrClass)FirstPrgLine+(StrClass)ExpressionVarName
            +(StrClass)"="+(StrClass)Expression+(StrClass)";"
            +(StrClass)LastPrgLine;
    #ifndef __OLD_INTERPRETE__
      // Retten des geparsten !!!
      const char * save = LastBufferPtr;
      Anweisungen * pAnweisungen = ObjInhalt.pAnweisungen;
      // Leere Struktur Simulieren ...
      LastBufferPtr=NULL;
      ObjInhalt.pAnweisungen=NULL;
      ObjInhalt.funktionen.FunktionenOff=1;
    #endif
      SetInputFileName("Calculator Buffer\n");
      if(StartParser((const char *)Buffer)==1)
      {
        #ifndef __OLD_INTERPRETE__
          ObjInhalt.Do();
        #endif
        result=GetReal(ExpressionVarName);
      }
    #ifndef __OLD_INTERPRETE__
      ObjInhalt.Do();
      LastBufferPtr=save;
      ObjInhalt.pAnweisungen=pAnweisungen;
      ObjInhalt.funktionen.FunktionenOff=0;
    #endif
  }
  return result;
}

/*------------------------------------------------------------------------------

    Konstruktor der Klasse Calculator
    --------------------------------------------
    void ::DeclareString(const char *VarName, const char *Wert)

    - Beschreibung
    --------------
    Deklarierung einer Variable mittels VariablenNamen und
    optionalen Wert

    - Parameter
    -----------
    E: VarName    - Variablen Name
    E: Wert        - Variablen Wert

    - allgemeine Ein- und Ausgabeeinheiten
    --------------------------------------
    dieselben des Parsers.


    - Returncode
    ------------
    keiner

    - Historie
    ----------
    6.7.1994 Sven Goethel

------------------------------------------------------------------------------*/
void Calculator::DeclareString(const char *VarName, const char *Wert)
{
  if( VarName!=NULL && strlen(VarName)>0  )
  {
    VariablenInhalt VarInhalt(0, STRING, 0, Wert);
    if( Vars.DefiniereVar ( VarName , VarInhalt) == 0 )
      ErrorState=VAR_DECLARE_ERR;
  }
}

/*------------------------------------------------------------------------------

    Konstruktor der Klasse Calculator
    --------------------------------------------
    void ::SetString(const char *VarName, const char *Wert)

    - Beschreibung
    --------------
    Setzen einer Variable mittels VariablenNamen und Wert

    - Parameter
    -----------
    E: VarName    - Variablen Name
    E: Wert        - Variablen Wert

    - allgemeine Ein- und Ausgabeeinheiten
    --------------------------------------
    dieselben des Parsers.


    - Returncode
    ------------
    keiner

    - Historie
    ----------
    6.7.1994 Sven Goethel

------------------------------------------------------------------------------*/
void Calculator::SetString(const char *VarName, const char *Wert)
{
  if( VarName!=NULL && strlen(VarName)>0  )
  {
    VariablenInhalt VarInhalt(0, STRING, 0, Wert);
    if( Vars.SetzeWert ( 0, VarName , VarInhalt) == 0 )
      ErrorState=VAR_SET_ERR;
  }
}

/*------------------------------------------------------------------------------

    Konstruktor der Klasse Calculator
    --------------------------------------------
    char *::GetString(const char *VarName)

    - Beschreibung
    --------------
    Holen einer Variable mittels VariablenNamen

    - Parameter
    -----------
    E: VarName    - Variablen Name

    - allgemeine Ein- und Ausgabeeinheiten
    --------------------------------------
    dieselben des Parsers.


    - Returncode
    ------------
    VariablenWert. Aktuell bis zum erneuerten Aufruf !

    - Historie
    ----------
    6.7.1994 Sven Goethel

------------------------------------------------------------------------------*/
char *Calculator::GetString(const char *VarName) 
{
  VariablenInhalt VarInhalt;
  static StrClass result;

  if( VarName!=NULL && strlen(VarName)>0  )
  {
    if(Vars.HoleWert(0, VarName, VarInhalt)==0)
    {
      ErrorState=VAR_GET_ERR;
    } else {
      if(VarInhalt.Type!=STRING) ErrorState=VAR_TYP_ERR;
      else {
        result=VarInhalt.strInhalt;
      }
    }
  }
  return (char *)(result.GetStrPtr());
}

/*------------------------------------------------------------------------------

    Konstruktor der Klasse Calculator
    --------------------------------------------
    char *::GetStringCopy(const char *VarName)

    - Beschreibung
    --------------
    Holen einer Variable mittels VariablenNamen
    GetStringCopy liefert eine Kopie mit malloc !!!!!!

    - Parameter
    -----------
    E: VarName    - Variablen Name

    - allgemeine Ein- und Ausgabeeinheiten
    --------------------------------------
    dieselben des Parsers.


    - Returncode
    ------------
    VariablenWert.
    Eine Kopie des VariablenWertes mit malloc !!!!!!

    - Historie
    ----------
    6.7.1994 Sven Goethel

------------------------------------------------------------------------------*/
char *Calculator::GetStringCopy(const char *VarName)
{
  char *result=NULL;
  const char *String=GetString(VarName);

  if( VarName!=NULL && strlen(VarName)>0  )
  {
    if(String!=NULL)
    {
      if ( ( result=(char *)malloc(strlen(String)+1) )
           == NULL
         )
      {
        ErrorState=STR_MALLOC_ERR;
      } else strcpy(result, String);
    }
  }
  return result;
}

void Calculator::SetInputFileName(const char *fn)
{
  Scanner.FileName=fn;
}

//#define __SHOW_CALC__
void Calculator::Interprete(FILE *input) 
{
  if(input!=NULL) 
  {
    StrClass leer="";
    if(Scanner.FileName==leer)
    {
      #ifndef NDEBUG
        Scanner.FileName ="\n  Bitte Filename mit Calculator::SetInputFileName";
        Scanner.FileName+="\n  immer vor dem Calculator::Interprete-Aufruf setzen.\n";
      #else
        Scanner.FileName="???";
      #endif
    }

    #ifndef __OLD_INTERPRETE__
      if(Scanner.FileName!=OldFileName)
      {
    #endif
        rewind(input);
        #ifdef __SHOW_CALC__
          #ifdef __OLD_INTERPRETE__
            fprintf(stderr, "%c", 248);
          #else
            fprintf(stderr, "%c", 253);
          #endif
        #endif
        StartParser(input);
        #ifdef __SHOW_CALC__
          fprintf(stderr, "|");
        #endif
    #ifndef __OLD_INTERPRETE__
        OldFileName=Scanner.FileName;
      }
      #ifdef __SHOW_CALC__
        fprintf(stderr, "(");
      #endif
      ObjInhalt.Do();
      #ifdef __SHOW_CALC__
        fprintf(stderr, ")");
      #endif
    #endif
    Scanner.FileName="";
  }
}
void Calculator::Interprete(const char *input)
{
  if(input!=NULL)
  {
    StrClass leer="";
    if(Scanner.FileName==leer)
    {
      #ifndef NDEBUG
        Scanner.FileName ="\n  Bitte Filename mit Calculator::SetInputFileName";
        Scanner.FileName+="\n  immer vor dem Calculator::Interprete-Aufruf setzen.\n";
      #else
        Scanner.FileName="???";
      #endif
    }
    #ifndef __OLD_INTERPRETE__
      if(input!=LastBufferPtr) 
    #endif
        StartParser(input);
    #ifndef __OLD_INTERPRETE__
        LastBufferPtr=input;
      ObjInhalt.Do();
    #endif
    Scanner.FileName="";
  }
}

/*------------------------------------------------------------------------------

    Konstruktor der Klasse Calculator
    --------------------------------------------
    Calculator::Error ::GetAndClearErrorState()

    - Beschreibung
    --------------
    Holen und loeschen der FehlerNummer/des Fehlerstatus

    - Parameter
    -----------
    keine

    - allgemeine Ein- und Ausgabeeinheiten
    --------------------------------------
    dieselben des Parsers.


    - Returncode
    ------------
    keiner

    - Historie
    ----------
    6.7.1994 Sven Goethel

------------------------------------------------------------------------------*/
Calculator::Error Calculator::GetAndClearErrorState()
{
  Error result=ErrorState;
  ErrorState=NO_ERR;
  return result;
}

/*------------------------------------------------------------------------------

    Konstruktor der Klasse Calculator
    --------------------------------------------
    const char *::GetErrorString(Calculator::Error ErrNum) const

    - Beschreibung
    --------------
    Holen des Fehlertextes mittels FehlerNummer

    - Parameter
    -----------
    E: ErrNum    - FehlerNummer

    - allgemeine Ein- und Ausgabeeinheiten
    --------------------------------------
    dieselben des Parsers.


    - Returncode
    ------------
    Zeiger auf Fehlertext

    - Historie
    ----------
    6.7.1994 Sven Goethel

------------------------------------------------------------------------------*/
const char *Calculator::GetErrorString(Calculator::Error ErrNum) const
{
  if(ErrNum>=WRONG_ERR_NUMBER) ErrNum=WRONG_ERR_NUMBER;
  return ErrorStrings[ErrNum];
}


/*------------------------------------------------------------------------------

    Konstruktor der Klasse Calculator
    --------------------------------------------
    void ::Reset()

    - Beschreibung
    --------------
    Zuruecksetzen der Variablen

    - Parameter
    -----------
    keine

    - allgemeine Ein- und Ausgabeeinheiten
    --------------------------------------
    dieselben des Parsers.


    - Returncode
    ------------
    Zeiger auf Fehlertext

    - Historie
    ----------
    6.7.1994 Sven Goethel

------------------------------------------------------------------------------*/
void Calculator::Reset()
{
  Vars.Dest();
  DeclareReal(ExpressionVarName); // ErrorState wird implizit gesetzt !!!
}

/*------------------------------------------------------------------------------

    Konstruktor der Klasse Calculator
    --------------------------------------------
    void ::AbortOnError(int Line)

    - Beschreibung
    --------------
    Wenn Fehler : Ausgabe des Fehlertextes, der uebergebenen Zeilennummer
                  und Abbruch

    - Parameter
    -----------
    E:Line        - Zeilennummer des Aufrufes um den Fehler zurueckzuverfolgen
                  i.d. Regel durch den Compiler vordefinierte
                  Variable __LINE__ !!

    - allgemeine Ein- und Ausgabeeinheiten
    --------------------------------------
    dieselben des Parsers.


    - Returncode
    ------------
    keiner

    - Historie
    ----------
    6.7.1994 Sven Goethel

------------------------------------------------------------------------------*/
void Calculator::AbortOnError(int Line)
{
  Error ErrNum=GetAndClearErrorState();

  if(ErrNum!=NO_ERR)
  {
    printf("<CALCULATOR ERROR:Line %d> %s !\n", Line, GetErrorString(ErrNum));
    exit(1);
  }
}


/*------------------------------------------------------------------------------

    Elementfunktion der Klasse Calculator
    --------------------------------------------
    int ::StartParser (FILE *input)

    - Beschreibung
    --------------
    Startet den Interpreter

    - Parameter
    -----------
    E: input    - die Eingabedatei

    - allgemeine Ein- und Ausgabeeinheiten
    --------------------------------------
    dieselben des Parsers.


    - Returncode
    ------------
    1 OK
    0 FEHLER

    - Historie
    ----------
    11.5.1994 Sven Goethel

------------------------------------------------------------------------------*/

int Calculator::StartParser (FILE *input)
{
  int result=0;

  #if YYDEBUG
    printf("...parser debug mode on !\n");
    #ifndef NO_PARSE_TEXT
      setdebug(1);
    #endif
  #endif

  if(result==0) {
    Scanner.setinput(input);
    Scanner.yy_reset();
    yyreset();
    result=yyparse(&Scanner); // Standard Typenkonversion
    if(result!=0) ErrorState=INTERPRETE_ERR;
  }
  return (result==0);
}

/*------------------------------------------------------------------------------

    Elementfunktion der Klasse Calculator
    --------------------------------------------
    int ::StartParser (const char *input)

    - Beschreibung
    --------------
    Startet den Interpreter

    - Parameter
    -----------
    E: input    - der Eingabebuffer

    - allgemeine Ein- und Ausgabeeinheiten
    --------------------------------------
    dieselben des Parsers.


    - Returncode
    ------------
    1 OK
    0 FEHLER

    - Historie
    ----------
    21.7.1994 Sven Goethel

------------------------------------------------------------------------------*/
int Calculator::StartParser (const char *input)
{
  int result=0;

  #if YYDEBUG
    printf("...parser debug mode on !\n");
    #ifndef NO_PARSE_TEXT
      setdebug(1);
    #endif
  #endif

  if(result==0 && strlen(input)>0 ) {
    Scanner.setinput(input);
    Scanner.yy_reset();
    yyreset();
    result=yyparse(&Scanner); // Standard Typenkonversion
    if(result!=0) ErrorState=INTERPRETE_ERR;
  }
  return (result==0);
}


