#ifndef yy_state_t
#define yy_state_t unsigned char
#endif
#define YYNEWLINE 10

// MKS LEX prototype scanner header
// Copyright 1991 by Mortice Kern Systems Inc.
// All rights reserved.

// You can define YY_PRESERVE to get System V UNIX lex compatibility,
//      if you need to change yytext[] in your user actions
// This is quite a bit slower, though, so the default is without
//
// Erweiterung am 6.7.1994 fuer den Calculator
// Protected-Klassen-Elemente wurden fuer den
// Interpreter eingefuegt. Selbige sind in der Klassendeklaration
// gekennzeichnet !!!
// Ebenfalls wurde yy_reset derart erweitert, so dass first_time
// wieder auf 1 gesetzt wird !!!
// Ebenfalls wurde eine vituelle Methode fuer fseek, yyfseek auf yyin,
// und ftell, yyftell auf yyin, implementiert !!!!! (21.7.1994)
//
/*
 *
 * $Log: calcscan.mks $
 * Revision 2.24  1995/04/07 11:35:08  Sven_Goethel_DFP
 * t_string wieder zurueck zu StrClass
 * Alles angepasst fuer Horts
 * ( aehh Horst . )
 * Wenn __SHOW_CALC__ definiert ist,
 * so werden Zeichen auf stderr gedruckt bei :
 *   PARSEN der DATEI
 *   INTERPRETIEREN der DATENSTRUKTUR
 * und alles fuer Calculator NEU und OLD
 * Revision 2.22  1995/04/05 07:45:38  Sven_Goethel_DFP
 * Alle Fehler behoben ..
 * getestet mit und ohne __OLD_INTERPRETE__ !
 * Fuer den neuen Interpreter sind Fehler-Message (FILE und ZEILENNUMMER)
 * hinzugekommen.
 * Fuer __OLD_INTERPRETE__ wurde ein BUG entfernt :
 *   bei return innerhalb verschachtelter Bloecke wurde der 
 *   Block zaehler nicht wieder zureuckgesetzt !!!
 * Revision 2.21  1995/04/05 02:00:18  Sven_Goethel_DFP
 * profekt-pflege nun mit KornShell-Sripts
 * Revision 2.2  1995/04/05 01:48:17  Sven_Goethel_DFP
 * __Mem_Free fehler beseitigt
 * Das MKS System hat um ein mittels new allokiertes Array zu deallokieren,
 * den delete operator verwendet.
 * Es muss aber der delete[] operator verwendet werden :
 *   int *p=new int[100];
 *   delete p; // Error !
 *   delete[] p; // Ok !
 * Revision 1.2  1994/12/31 00:59:05  Sven_Goethel
 * non virtual for WATCOM
 * Revision 1.1  1994/12/16 16:52:11  Sven_Goethel
 * Initial revision
 *
*/

#ifndef _YY_SCAN_H
#define _YY_SCAN_H

#include <stdio.h>              // uses printf(), et cetera
#include <stdarg.h>             // uses va_list
#include <stdlib.h>             // uses exit()
#include <string.h>             // uses memmove()
#include <strclass.h>          

const int BufferSize = 200;

class yy_scan {
protected:
	yy_state_t * state;             // state buffer
	int     size;                   // length of state buffer
	int     mustfree;               // set if space is allocated
	int     yy_end;                 // end of pushback
	int     yy_start;               // start state
	int     yy_lastc;               // previous char
#ifndef YY_PRESERVE
				// efficient default push-back scheme
	char save;              // saved yytext[yyleng]
#else
				// slower push-back for yytext mungers
	char *save;             // saved yytext[]
	char *push;
#endif
	// Erweiterung fuer den FxmConverterInterpreter <START> !
	//
	int first_time;
	char Buffer[BufferSize];
	// PassNo=0 : Im ersten Pass die Lokation der FxmObjekte Orten ...
	// PassNo=1 : Im zweiten Pass wirklich Parsen ...
	int PassNo;
	//
	// Erweiterung fuer den FxmConverterInterpreter <ENDE> !

public:
	char * yytext;          // yytext text buffer
	FILE   *yyin;                   // input stream
	FILE   *yyout;                  // output stream
	int     yylineno;               // line number
	int     yyleng;                 // yytext token length
	// filename information <START>
	StrClass FileName;
	//                      <ENDE>

	yy_scan(int = 100);     // constructor for this scanner
			// default token & pushback size is 100 bytes
	yy_scan(int, char*, char*, yy_state_t*);
				// constructor when tables are given

	virtual ~yy_scan();             // destructor

	int     yylex();                // begin a scan

	virtual int     yygetc() {      // scanner source of input characters
		return getc(yyin);
	}

	// ERWEITERUNGEN :
	// scanner source of manipulate FilePointer
	virtual int     yyfseek(long offset, int whence) {
		return fseek(yyin, offset, whence);
	}
	// scanner source of getting FilePointer
	virtual long yyftell() {
		return ftell(yyin);
	}

	virtual int     yywrap() { return 1; }  // EOF processing

#ifndef __WATCOMC__
	virtual
#endif
	void    yyerror(char *,...);    // print error message

	virtual void    output(int c) { putc(c, yyout); }

	virtual void    YY_FATAL(char * msg) {  // print message and stop
		yyerror(msg); exit(1);
	}
	virtual void    ECHO() {                // print matched input
		fputs((const char *) yytext, yyout);
	}
	int     input();                // user-callable get-input
	int     unput(int c);           // user-callable unput character
	void    yy_reset();             // reset scanner
	void    setinput(FILE * in) {           // switch input streams
		yyin = in;
	}
	void    setoutput(FILE * out) { // switch output
		yyout = out;
	}
	void    NLSTATE() { yy_lastc = YYNEWLINE; }
	void    YY_INIT() {
		yy_start = 0;
		yyleng = yy_end = 0;
		yy_lastc = YYNEWLINE;
	}
	void    YY_USER() {             // set up yytext for user
#ifndef YY_PRESERVE
		save = yytext[yyleng];
#else
		size_t n = yy_end - yyleng;
		push = save+size - n;
		if (n > 0)
			memmove(push, yytext+yyleng, n);
#endif
		yytext[yyleng] = 0;
	}
	void YY_SCANNER() {             // set up yytext for scanner
#ifndef YY_PRESERVE
		yytext[yyleng] = save;
#else
		size_t n = save+size - push;
		if (n > 0)
			memmove(yytext+yyleng, push, n);
		yy_end = yyleng + n;
#endif
	}
	void    yyless(int n) {         // trim input to 'n' bytes
		if (n >= 0 && n <= yy_end) {
			YY_SCANNER();
			yyleng = n;
			YY_USER();
		}
	}
	void    yycomment(char *const mat); // skip comment input
	int     yymapch(int delim, int escape); // map C escapes
} ;

#endif

